#include <linux/module.h>//for kernel module
#include <linux/kernel.h>//for kernel module
#include <linux/init.h>//for kernel module
#include <linux/slab.h>//for kmalloc
#include <linux/time.h>//for current_kernel_time
#include <linux/proc_fs.h>//for proc
#include <linux/sched.h>//for schedule
#include <asm/uaccess.h>//for copy_from_user
#include "aes.h"//for OpenSSL-extracted aes

//global kernel variables
static unsigned char* aes_key;
static AES_KEY aes_encryptionkeyschedule;
static AES_KEY aes_decryptionkeyschedule;
static struct proc_dir_entry* aesprocfile;
static char* aesprocbuffer;
static unsigned long aesprocbuflen = 0;

//this function is executed when you try to read from /proc/aestest
static int aestest_read_proc(char* buffer, char** start, off_t offset, int size, int* eof, void* data){
	u32 i;
	u32 len;
	struct timespec starttime;
	struct timespec endtime;
	int bufsize = 512*2;
	char* buf;
	unsigned char* aes_ivec;
	unsigned char* aes_ivec2;
	unsigned int aes_cbcoffset;

	printk(KERN_INFO "read_proc in aestest module...\n");
	printk("AES_BLOCK_SIZE: %d\n",AES_BLOCK_SIZE);//should be 16
	//set arbitrary encryption key for testing
	if(aes_key == NULL){
		aes_key=(unsigned char*)kmalloc(32,GFP_ATOMIC);
		for(i=0; i < 32; i++) aes_key[i] = 'k';
		printk("key: %s\n",aes_key);
		if(AES_set_encrypt_key(aes_key,256,&(aes_encryptionkeyschedule))!=0)
			printk("Error: AES_set_encrypt_key\n");
		if(AES_set_decrypt_key(aes_key,256,&(aes_decryptionkeyschedule))!=0)
			printk("Error: AES_set_decrypt_key\n");
	}
	//set arbitrary CBC initialization vector
	aes_ivec=(unsigned char*)kmalloc(AES_BLOCK_SIZE,GFP_ATOMIC);
	aes_ivec2=(unsigned char*)kmalloc(AES_BLOCK_SIZE,GFP_ATOMIC);
	for(i=0; i < AES_BLOCK_SIZE; i++) aes_ivec[i] = 'i';
	memcpy(aes_ivec2,aes_ivec,AES_BLOCK_SIZE);

	//buffer starts out as a-z looped for bufsize bytes
	buf = (char*)kmalloc(bufsize,GFP_ATOMIC);
	for(i=0;i < bufsize; i++) buf[i] = (unsigned char)(i%26+97);

	//print plaintext version
	printk("buf: %s\n",buf);

	//set start time
	starttime = current_kernel_time();
	
	//encrypt buffer 1 million times with the same key
	for(i=0; i < 1000000; i++){
		//perform aes-256-CBC on 512 bytes at a time using the same iv
		for(aes_cbcoffset=0;aes_cbcoffset<bufsize/512;aes_cbcoffset++){
			memcpy(aes_ivec,aes_ivec2,AES_BLOCK_SIZE);
			AES_cbc_encrypt(buf+aes_cbcoffset*512,buf+aes_cbcoffset*512,512,&(aes_encryptionkeyschedule),aes_ivec,AES_ENCRYPT);
		}
		if(i%1000==0) schedule();//don't hog the cpu
	}

	//print first char of cyphertext version
	printk("buf: %c\n",buf[0]);

	//decrypt buffer 1 million times with the same key
	for(i=0; i < 1000000; i++){
		//perform aes-256-CBC on 512 bytes at a time using the same iv
		for(aes_cbcoffset=0;aes_cbcoffset<bufsize/512;aes_cbcoffset++){
			memcpy(aes_ivec,aes_ivec2,AES_BLOCK_SIZE);
			AES_cbc_encrypt(buf+aes_cbcoffset*512,buf+aes_cbcoffset*512,512,&(aes_decryptionkeyschedule),aes_ivec,AES_DECRYPT);
		}
		if(i%1000==0)
			schedule();//don't hog the cpu
	}

	//print plaintext version, which matches the original
	printk("buf: %s\n",buf);

	//determine end time and calculate stats
	endtime = current_kernel_time();
	printk("elapsed: %d\n",endtime.tv_sec-starttime.tv_sec);
	printk("MB/s: %d\n",(2*bufsize)/(endtime.tv_sec-starttime.tv_sec));

	kfree(buf);
	kfree(aes_ivec);
	kfree(aes_ivec2);

	len = strlen(aesprocbuffer);
	if(size < len) return -EINVAL;
	if(offset != 0) return 0;
	strcpy(buffer,aesprocbuffer);
	*eof = 1;
	return len;
}//end read_proc

//this function is called when you try to write to /proc/aestest
static int aestest_write_proc(struct file *file, const char* buffer, unsigned long count, void* data){
	aesprocbuflen = count;
	if(aesprocbuflen > 512){
		aesprocbuflen = 512;
	}
	if(copy_from_user(aesprocbuffer,buffer,aesprocbuflen)){
		return -EFAULT;
	}
	aesprocbuffer[aesprocbuflen] = '\0';
	printk("module received: %s\n",aesprocbuffer);
	return aesprocbuflen;
}//end write_proc

//initialize kernel module and create /proc/aestest
static int __init aesteststart(void){
	printk(KERN_INFO "Loading aestest module...\n");
	aesprocbuffer = (char*)kmalloc(512,GFP_ATOMIC);
	aesprocfile = create_proc_entry("aestest",0644,NULL);
	if(aesprocfile == NULL){
		remove_proc_entry("aestest",NULL);
		printk("coult not initialize /proc/aestest\n");
	}
	aesprocfile->read_proc = aestest_read_proc;
	aesprocfile->write_proc = aestest_write_proc;
	aesprocfile->owner = THIS_MODULE;
	aesprocfile->mode = S_IFREG|S_IRUGO;
	aesprocfile->uid = 0;
	aesprocfile->gid = 0;
	aesprocfile->size = 37;
	printk("created /proc/aestest\n");
	printk(KERN_INFO "Hello from aestest module\n");
	return 0;
}//end __init aesteststart

//remove kernel module and /proc/aestest
static void __exit aestestend(void){
	if(aes_key != NULL)kfree(aes_key);
	kfree(aesprocbuffer);
	remove_proc_entry("aestest",NULL);
	printk(KERN_INFO "Goodbye from aestest module\n");
}//end __exit aestestend

module_init(aesteststart);
module_exit(aestestend);
MODULE_DESCRIPTION("AES test module");
