#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
//defined in kernel and used in aes.h, so define here
typedef unsigned int u32;
typedef unsigned short u16;
typedef unsigned char u8;
#include "aes.h"

//global variables match kernel module
static unsigned char* aes_key;
static AES_KEY aes_encryptionkeyschedule;
static AES_KEY aes_decryptionkeyschedule;

int main(void){
	int i;
	time_t starttime;
	time_t endtime;
	int bufsize = 512*2;
	char* buf;
	unsigned char* aes_ivec;
	unsigned char* aes_ivec2;
	unsigned int aes_cbcoffset;

	printf("main in aestest user process...\n");
	printf("AES_BLOCK_SIZE: %d\n",AES_BLOCK_SIZE);//should be 16	
	//set arbitrary encryption key for testing
	if(aes_key == NULL){
		aes_key=(unsigned char*)malloc(32);
		for(i=0; i < 32; i++) aes_key[i] = 'k';
		printf("key: %s\n",aes_key);
		if(AES_set_encrypt_key(aes_key,256,&(aes_encryptionkeyschedule))!=0)
			printf("Error: AES_set_encrypt_key\n");
		if(AES_set_decrypt_key(aes_key,256,&(aes_decryptionkeyschedule))!=0)
			printf("Error: AES_set_decrypt_key\n");
	}
	//set arbitrary CBC initialization vector
	aes_ivec=(unsigned char*)malloc(AES_BLOCK_SIZE);
	aes_ivec2=(unsigned char*)malloc(AES_BLOCK_SIZE);
	for(i=0; i < AES_BLOCK_SIZE; i++) aes_ivec[i] = 'r';
	memcpy(aes_ivec2,aes_ivec,AES_BLOCK_SIZE);

	//buffer starts out as a-z looped for bufsize bytes
	buf = (char*)malloc(bufsize);
	for(i=0;i < bufsize; i++) buf[i] = (unsigned char)(i%26+97);

	//print plaintext version
	printf("buf: %s\n",buf);

	//set start time
	time(&starttime);
	
	//encrypt buffer 1 million times with the same key
	for(i=0; i < 1000000; i++){
		//perform aes-256-CBC on 512 bytes at a time using the same iv
		for(aes_cbcoffset=0;aes_cbcoffset<bufsize/512;aes_cbcoffset++){
			memcpy(aes_ivec,aes_ivec2,AES_BLOCK_SIZE);
			AES_cbc_encrypt(buf+aes_cbcoffset*512,buf+aes_cbcoffset*512,512,&(aes_encryptionkeyschedule),aes_ivec,AES_ENCRYPT);
		}
	}

	//print first char of cyphertext version
	printf("buf: %c\n",buf[0]);

	//decrypt buffer 1 million times with the same key
	for(i=0; i < 1000000; i++){
		//perform aes-256-CBC on 512 bytes at a time using the same iv
		for(aes_cbcoffset=0;aes_cbcoffset<bufsize/512;aes_cbcoffset++){
			memcpy(aes_ivec,aes_ivec2,AES_BLOCK_SIZE);
			AES_cbc_encrypt(buf+aes_cbcoffset*512,buf+aes_cbcoffset*512,512,&(aes_decryptionkeyschedule),aes_ivec,AES_DECRYPT);
		}
	}

	//print plaintext version, which matches the original
	printf("buf: %s\n",buf);

	//determine end time and calculate stats
	time(&endtime);
	printf("elapsed: %d\n",endtime-starttime);
	printf("MB/s: %d\n",(2*bufsize)/(endtime-starttime));

	free(buf);
	free(aes_ivec);
	free(aes_ivec2);
	free(aes_key);
	return 0;
}//end main
